/****************************************************************************
 ** Makou Reactor Final Fantasy VII Field Script Editor
 ** Copyright (C) 2009-2012 Arzel Jrme <myst6re@gmail.com>
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ****************************************************************************/
#ifndef ISOARCHIVE_H
#define ISOARCHIVE_H

#include <QtCore>

#define MAX_ISO_READ			10000
#define MAX_FILENAME_LENGTH		207
#define SEPARATOR_1				'\x2E' // .
#define SEPARATOR_2				'\x3B' // ;

//d-characters : [\w ]
//a-characters : [!"%&'\(\)*+,\-\./\w:;<=>? ] (+ SP)

typedef struct{
	char year[4];
	char month[2];
	char day[2];
	char hour[2];
	char minute[2];
	char second[2];
	char millis[2];
	qint8 GMT;
} IsoTime;

typedef struct pathTbl{
	quint8 length_di;
	quint8 extended_attr_record_length;
	quint32 location_extent;
	quint16 parent_directory_number;
	QString name;
	quint32 position;
} PathTable;

typedef struct{
	quint32 location_extent;// little endian //first sector in the directory
	quint32 location_extent2;// big endian
	quint32 data_length;// little endian
	quint32 data_length2;// big endian
	quint8 year;// since 1900
	quint8 month;
	quint8 day;
	quint8 hour;
	quint8 minute;
	quint8 second;
	qint8 GMT;
	quint8 file_flags;// Multi-Extent/Reserved/Reserved/Protection/Record/Associated File/Directory/Existence
	quint8 file_unit_size;
	quint8 interleave_grap_size;
	quint16 volume_sequence_number;// little endian
	quint16 volume_sequence_number2;// big endian
	quint8 length_fi;
} DirectoryRecordHead;

typedef struct dirRec{
	quint8 length_dr;
	quint8 extended_attr_record_length;
	DirectoryRecordHead drh;
	QString name;
	QString version;
	// padding;
	// system use
	QList<struct dirRec> directories;
	quint32 position;
} DirectoryRecord;

typedef struct{
	quint8 type;// 0x01 or 0xff
	quint8 id[5];// CD001
	quint8 version;// 1 = international standard
	quint8 unused1;
	quint8 system_id[32];
	quint8 volume_id[32];
	quint8 unused2[8];
	quint32 volume_space_size;// little endian // number of sectors
	quint32 volume_space_size2;// big endian
	quint8 unused3[32];
	quint16 volume_set_size;// little endian
	quint16 volume_set_size2;// big endian
	quint16 volume_sequence_number;// little endian
	quint16 volume_sequence_number2;// big endian
	quint16 logical_block_size;// little endian
	quint16 logical_block_size2;// big endian
	quint32 path_table_size;// little endian
	quint32 path_table_size2;// big endian
	quint32 type_path_table;// little endian
	quint32 opt_type_path_table;// little endian
	quint32 type_path_table2;// big endian
	quint32 opt_type_path_table2;// big endian
} VolumeDescriptor1;

typedef struct{
	quint8 volume_set_id[128];
	quint8 publisher_id[128];
	quint8 preparer_id[128];
	quint8 application_id[128];
	quint8 copyright_file_id[37];
	quint8 abstract_file_id[37];
	quint8 bibliographic_file_id[37];
	IsoTime creation_date;
	IsoTime modification_date;
	IsoTime expiration_date;
	IsoTime effective_date;
	quint8 file_structure_version;
	quint8 unused4;
	quint8 application_data[512];
	quint8 unused5[653];
} VolumeDescriptor2;

typedef struct{
	VolumeDescriptor1 vd1;
	DirectoryRecord dr;
	VolumeDescriptor2 vd2;
} VolumeDescriptor;

class IsoArchive
{
public:
	IsoArchive(const QString &name);
	~IsoArchive();

	quint8 open(QIODevice::OpenMode mode);
	void close();
	qint64 pos() const;
	static qint64 pos(QFile *file);
	qint64 rawPos() const;
	bool seek(qint64 off);
	static bool seek(qint64 off, QFile *file);
	bool rawSeek(qint64 off);
	qint64 size() const;
	static qint64 size(QFile *file);
	qint64 rawSize() const;

	QString fileName();

	qint64 read(char *data, qint64 maxSize);
	QByteArray read(qint64 maxSize);
	static qint64 read(char *data, qint64 maxSize, QFile *file);
	static QByteArray read(qint64 maxSize, QFile *file);
	qint64 rawRead(char *data, qint64 maxSize);
	QByteArray rawRead(qint64 maxSize);

	qint64 write(char *data, qint64 maxSize);
	qint64 write(QByteArray &byteArray);
	static qint64 write(char *data, qint64 maxSize, QFile *file);
	static qint64 write(QByteArray &byteArray, QFile *file);
	qint64 rawWrite(char *data, qint64 maxSize);
	qint64 rawWrite(QByteArray &byteArray);

	QByteArray file(QString path, quint32 maxSize=0);
	bool extract(QString path, QString destination, quint32 maxSize=0);
	void extractAll(QString destination);
	bool seekToFile(QString path);
	qint32 diffCountSectors(QString path, quint32 newSize);
	bool infosFile(QString path, quint32 &position, quint32 &size);
	
	static QByteArray int2Header(quint32 id);
	static QByteArray buildHeader(quint32 sector, quint8 type, quint8 mode=2);
	static QByteArray buildFooter(quint32 sector);
	
	QByteArray sector(quint32 num, quint16 maxSize=2048);
	QByteArray sectorHeader(quint32 num);
	QByteArray sectorFooter(quint32 num);
	quint32 sectorCount();
	bool seekToSector(quint32 num);
	
	QList<DirectoryRecord> getDirectories();
	bool getDirectoryRecord(QString path, DirectoryRecord &fileRecord);
	QList<PathTable> getPathTables1a();
	QList<PathTable> getPathTables1b();
	QList<PathTable> getPathTables2a();
	QList<PathTable> getPathTables2b();
private:
	int indexOfDirectoryRecord(QString name, QList<DirectoryRecord> directories);
	VolumeDescriptor volumeDescriptor(quint8 num=0);
	QList<DirectoryRecord> directoryRecord(quint32 sector, quint32 dataSize=2048);
	QList<PathTable> pathTable(quint32 sector, quint32 dataSize=2048);
	
	static qint64 isoPos(qint64 pos);
	static qint64 filePos(qint64 pos);
	static QString isoTimeToString(IsoTime time);
	static QString volumeDescriptorToString(VolumeDescriptor vd);
	static QString directoryRecordToString(DirectoryRecord dr);
	static QString pathTableToString(PathTable pathTable, bool bigEndian=false);
	
	void _extractAll(QString destination, QList<DirectoryRecord> drs, QString currentInternalDir=QString());
	
	static quint8 hex2Dec(quint8 hex);
	static quint8 dec2Hex(quint8 dec);
	
	QFile _file;
	VolumeDescriptor volume;
	QList<DirectoryRecord> directories;
	/*QList<PathTable> pathTables1a;
	QList<PathTable> pathTables1b;
	QList<PathTable> pathTables2a;
	QList<PathTable> pathTables2b;*/
	QList<quint32> dirVisisted;
};

#endif // ISOARCHIVE_H
